/* This little tool listens on a port, and connects the first connection
to the second connection.  Thus, two machines (A and C) inside different
firewalls can telnet out to this broker, running on a third unrestricted
machine (B), and be connected.  A --> B <-- C  ==>  A <--> B <--> C.  */

#include "tcpbroker.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

int Sock[2];                    /* The clients are on these sockets. */
int Running;                    /* Flag that we are connected. */
int Maxfd;                      /* for select() */

#define MSGBUFSIZE (8 * 1024)   /* Max size of messages the server allows. */

int wait_for_input();
int authorized();
void copy_input(int, int);
void cleanup(int);

int main(int argc, char **argv)
{
	int sock, s;
	unsigned int i, j;
	struct sockaddr_in sock_name;

	if (argc != 2) {
		fprintf(stderr, "Usage: %s portnumber\n", argv[0]);
		exit(1);
	}

	Running = FALSE;

	/* Open a socket. */

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	/* Assign the socket's address and port. */

	memset(&sock_name, 0, sizeof(sock_name));
	sock_name.sin_family = AF_INET;
	sock_name.sin_port = htons(atoi(argv[1]));
	sock_name.sin_addr.s_addr = INADDR_ANY;

	/* Bind the address to the socket. */

	if (bind(sock, (struct sockaddr *)&sock_name, sizeof(sock_name))) {
		perror("bind");
		cleanup(sock);
	}

	/* Set this socket to listen for connections. */

	if (listen(sock, 5)) {
		perror("listen");
		cleanup(sock);
	}

	/* Wait for two connections. */

	fprintf(stderr, "waiting for connections\n");
	j = sizeof(sock_name);
	if ((Sock[0] = accept(sock, (struct sockaddr *)&sock_name, &j)) < 0) {
		perror("accept");
		cleanup(sock);
	}
	i = ntohl(sock_name.sin_addr.s_addr);
	fprintf(stderr, "connected A: %d.%d.%d.%d/%d\n",
		(i >> 24) & 0xFF, (i >> 16) & 0xFF,
		(i >> 8) & 0xFF, i & 0xFF,
		ntohs(sock_name.sin_port));
	if ((Sock[1] = accept(sock, (struct sockaddr *)&sock_name, &j)) < 0) {
		perror("accept");
		close(Sock[0]);
		cleanup(sock);
	}
	i = ntohl(sock_name.sin_addr.s_addr);
	fprintf(stderr, "connected C: %d.%d.%d.%d/%d\n",
		(i >> 24) & 0xFF, (i >> 16) & 0xFF,
		(i >> 8) & 0xFF, i & 0xFF,
		ntohs(sock_name.sin_port));
	j = Sock[0] > Sock[1] ? Sock[0] : Sock[1];
	Maxfd = j + 1;

	/* We have a potential connection.  One side must now authenticate. */

	if (authorized()) {
		fprintf(stderr, "connection authorized\n");

		/* Start copying data. */

		Running = TRUE;
		while (Running) {

			/* Watch for input on either socket. */

			s = wait_for_input();
			if (s >= 0) {
				copy_input(Sock[s], Sock[1 - s]);
			}
		}
	}

	/* Shut down the sockets. */

	shutdown(Sock[0], 2);
	close(Sock[0]);
	shutdown(Sock[1], 2);
	close(Sock[1]);
	close(sock);

	return 0;
}

/* Return a socket that has input ready.  There might be more than one, in
which case we save it for the next call.  If we get a timeout, return -1. */

int wait_for_input()
{
	int i;
	fd_set fds;
	struct timeval tv;
	static int save1 = 0;

	if (save1) {
		save1 = 0;
		return 1;
	}

	FD_ZERO(&fds);
	FD_SET(Sock[0], &fds);
	FD_SET(Sock[1], &fds);
	tv.tv_sec = 60;
	tv.tv_usec = 0;
	if (select(Maxfd, &fds, NULL, NULL, &tv) < 0) {
		return -1;
	}

	i = -1;
	if (FD_ISSET(Sock[0], &fds)) {
		i = 0;
	}
	if (FD_ISSET(Sock[1], &fds)) {
		if (i < 0) {
			i = 1;
		} else {
			save1 = 1;
		}
	}

	return i;
}

void timeout(int s)
{
	/* I suppose we could do something more drastic here, like
	shutting down the sockets and exiting... */

	Running = FALSE;
}

/* Authorize the connection.  We only need authentication from one side,
but we don't know which side.  So first we'll wait for whichever side sends
us the magic cookie.  The other side isn't required to do anything. */

int authorized()
{
	int i, len, s, suff;
	char peekbuf[TSRP_COOKIE_LEN];

	/* Wait for a cookie.  Abort on timeout, recv() error, or the
	case where we already have more than enough data from both sides. */

	signal(SIGALRM, timeout);
	alarm(30);
	Running = TRUE;

	suff = 0;
	while (Running) {
		s = wait_for_input();
		if (s < 0) {
			return FALSE;
		}
		len = recv(Sock[s], peekbuf, sizeof(peekbuf), MSG_PEEK);
		if (len <= 0) {
			return FALSE;
		}
		if (len >= TSRP_COOKIE_LEN) {
			if (memcmp(peekbuf, TSRP_COOKIE, TSRP_COOKIE_LEN) == 0) {
				break;
			}
			suff |= (1 << s);
			if (suff == 3) {
				return FALSE;
			}
		}
	}
	if (!Running) {
		return FALSE;
	}

	/* We got a cookie on s.  Eat it, reply, and pass the socket on to
	the authentication layer. */

	len = recv(Sock[s], peekbuf, sizeof(peekbuf), 0);
	if (len <= 0) {
		return FALSE;
	}
	send(Sock[s], TSRP_COOKIE, TSRP_COOKIE_LEN, 0);
	i = tsrp_server_authenticate(Sock[s], NULL);
	if (!Running) {
		return FALSE;
	}

	/* All done before the alarm went off. */

	alarm(0);
	Running = FALSE;
	return i;
}

/* If there is any input on from, send it to to. */

char Msgbuf[MSGBUFSIZE];

void copy_input(int from, int to)
{
	int i;

	if ((i = recv(from, Msgbuf, MSGBUFSIZE, 0)) < 0) {
		Running = FALSE;
		perror("recv");
	} else {

		/* If i == 0, it means the other end went away. */

		if (i == 0) {
			Running = FALSE;
		} else {

			/* We got a message from the socket.  Send it out. */

			send(to, Msgbuf, i, 0);
		}
	}
}

void cleanup(int sock)
{
	close(sock);
	exit(1);
}
